/*  Copyright (C) 2006, 2007 William McCune

    This file is part of the LADR Deduction Library.

    The LADR Deduction Library is free software; you can redistribute it
    and/or modify it under the terms of the GNU General Public License,
    version 2.

    The LADR Deduction Library is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the LADR Deduction Library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include "semantics.h"

/* Private definitions and types */

static Plist Compiled_interps = NULL;
static int Max_domain_size = 0;

/*************
 *
 *   init_semantics()
 *
 *************/

/* DOCUMENTATION
Take a list of terms that represent interpretations,
compile them, and store them locally.
*/

/* PUBLIC */
void init_semantics(Plist interp_terms)
{
  Plist p;
  
  for (p = interp_terms; p; p = p->next) {
    Interp a = compile_interp(p->v, FALSE);
    Max_domain_size = IMAX(Max_domain_size, interp_size(a));
    Compiled_interps = plist_prepend(Compiled_interps, a);
  }
  Compiled_interps = reverse_plist(Compiled_interps);
}  /* init_semantics */

/*************
 *
 *   eval_in_interps()
 *
 *************/

/* DOCUMENTATION
Evaluate a clause c in the compiled interpretations.  The
parameter "type" should be "false_in_all" or "false_in_some".
<p>
If evaluation would be too expensive (eval_limit), or if
a the clause contains an uninterpreted symbol, the clause
gets the valuse TRUE for that interpretation.
<p>
Finally, if c->semantics gets FALSE, and there are some
interpretations, the clause gets the attribute "false"
(which is intended to be used only for printing).
*/

/* PUBLIC */
void eval_in_interps(Topform c, int label_att, char *type, int eval_limit)
{
  if (Compiled_interps) {

    BOOL false_in_all = str_ident(type, "false_in_all");
    int num_vars = number_of_variables(c->literals);
    Plist p;      

    if (!c->normal_vars)
      renumber_variables(c, MAX_VARS);

    /* This is messy, because both cases (false_in_all, false_in_some)
       are handled in one loop.  Start by assuming we'll make it through
       all interps.  If we find we should stop along the way, change the
       result and break.
    */

    c->semantics = (false_in_all ? FALSE : TRUE);

    for (p = Compiled_interps; p; p = p->next) {
      Interp x = p->v;
      BOOL eval_ok, val, fail;
      if (eval_limit == -1)
	eval_ok = TRUE;
      else {
	int evals_required = int_power(interp_size(x), num_vars);
	eval_ok = evals_required <= eval_limit;
      }

      if (!eval_ok)
	val = TRUE;
      else if (!evaluable_clause(c, x)) {
	val = TRUE;
	printf("\n%% Clause contains symbol not in interpretation:\n");
	printf("%% "); fwrite_clause(stdout, c, CL_FORM_STD);
      }
      else if (eval_literals(c->literals, x) == TRUE)
	val = TRUE;
      else
	val = FALSE;

      fail = (false_in_all ? val == TRUE : val == FALSE);
      if (fail) {
	c->semantics = (false_in_all ? TRUE : FALSE);
	break;
      }
    }

    if (c->semantics == FALSE)
      c->attributes = set_string_attribute(c->attributes, label_att, "false");
  }

  else {
    /* There are no interps, so use default interp: positive literals TRUE. */
    c->semantics = !negative_clause(c->literals);
  }
}  /* eval_in_interps */

/*************
 *
 *   update_semantics_new_constant()
 *
 *************/

/* DOCUMENTATION
A new constant, say c_0, has just been introduced with
a clause, say f(x,x) = c_0.
This routine adds c_0 to all of the compiled interpretations.
It is given the same value as f(0,0).
*/

/* PUBLIC */
void update_semantics_new_constant(Topform c)
{
  Term alpha = ARG(c->literals->atom, 0);
  Term beta = ARG(c->literals->atom, 1);
  Plist p;
  int n = biggest_variable(alpha);
  int *vals = malloc((n+1) * sizeof(int));
  int i;

  for (i = 0; i <= n; i++)
    vals[i] = 0;

  for (p = Compiled_interps, i = 1; p; p = p->next, i++) {
    Interp x = p->v;
    int val = eval_term_ground(alpha, x, vals);
    update_interp_with_constant(x, beta, val);
    printf("NOTE: updating interpretation %d: %s=%d.\n",
	   i, sn_to_str(SYMNUM(beta)), val);
  }

  free(vals);
}  /* update_semantics_new_constant */
